from collections import OrderedDict
from enum import Enum
import re
from typing import Any, AnyStr, Dict, List, NamedTuple


KEY_SEP_CHAR = "."
KEY_ARRAY = "[]"


def escape_key(key: AnyStr, sep: str = None) -> AnyStr:
    sep = sep or KEY_SEP_CHAR
    escape_char = rf"\\{sep}"
    return key.replace(sep, escape_char)


def unescape_key(key: AnyStr, sep: str = None) -> AnyStr:
    sep = sep or KEY_SEP_CHAR
    escape_char = rf"\\{sep}"
    return key.replace(escape_char, sep)


def split_key(key: AnyStr, sep: str = None) -> AnyStr:
    sep = sep or KEY_SEP_CHAR
    s = rf"(?<!\\)\{sep}"
    p = re.compile(s)
    return re.split(p, key)


def join_key(parts: List[AnyStr], sep: str = None) -> AnyStr:
    sep = sep or KEY_SEP_CHAR
    return sep.join(parts)


class KV_Pair(NamedTuple):
    key: str
    value: Any


class FlatListType(Enum):
    SKIP = 0
    WITH_INDEX = 1
    WITHOUT_INDEX = 2
    WITH_SMART_INDEX = 3


def flatten_as_dict(
    data,
    sep=KEY_SEP_CHAR,
    flat_list=False,
    use_ordered_dict=True,
) -> Dict:
    out: Dict = OrderedDict() if use_ordered_dict else dict()
    sep_len = len(sep)

    def flatten(obj, name=""):
        if isinstance(obj, dict):
            for attr in obj:
                flatten(obj[attr], f"{name}{escape_key(attr)}{sep}")
        elif flat_list and isinstance(obj, list):
            parent_name = f"{name[:-sep_len]}{KEY_ARRAY}{sep}"
            # use OrderedDict to retain the order, should not use index,
            # the index is considered as a "key" to differentiate:
            for i in range(len(obj)):
                key = f"{parent_name}{i}{sep}"
                flatten(obj[i], key)
        else:
            key = name[:-sep_len]
            out[key] = obj

    if data:
        flatten(data)
    # @ret:
    return out


def flatten_as_list(
    data, sep=KEY_SEP_CHAR, flat_list_type: FlatListType = FlatListType.SKIP
) -> List[KV_Pair]:
    out: List = []
    out_keys = {}
    out_ref_keys = {}
    sep_len = len(sep)
    flat_list = flat_list_type in [
        FlatListType.WITH_INDEX,
        FlatListType.WITHOUT_INDEX,
        FlatListType.WITH_SMART_INDEX,
    ]

    def flatten(obj, name="", ref_name=None):
        if isinstance(obj, dict):
            for attr in obj:
                key = f"{name}{escape_key(attr)}{sep}"
                # print("1.0", [key, ref_name])
                flatten(obj[attr], key, ref_name)
        elif flat_list and isinstance(obj, list):
            # if ref_name is existing, means set the value to the last item of the array,
            # thus, the parent must refer "-1" (instead of "#" -> append):
            if ref_name is None or ref_name not in out_keys:
                parent_name = f"{name[:-sep_len]}{KEY_ARRAY}{sep}"
            else:
                t_key = out_ref_keys[ref_name]
                t_len = len(t_key) - 1  # -1 : exclude the sign "-".
                t_key += name[t_len:]
                parent_name = f"{t_key[:-sep_len]}{KEY_ARRAY}{sep}"
            # use the ordered in out-List to retain the order, should not use index,
            # the index is considered as a "key" to differentiate:
            if flat_list_type == FlatListType.WITH_INDEX:
                # all "indexes" will be added to output keys:
                for i in range(len(obj)):
                    key = f"{parent_name}{i}{sep}"
                    # print("2.1", [key, ref_name])
                    flatten(obj[i], key, ref_name)
            elif flat_list_type == FlatListType.WITHOUT_INDEX:
                # all "indexes" will be replaced by "#":
                for i in range(len(obj)):
                    key = f"{parent_name}#{sep}"
                    # print("2.2", [key, ref_name])
                    flatten(obj[i], key, ref_name)
            else:  # WITH_SMART_INDEX:
                # all "indexes" will be replaced by (same length):
                #  - "#" if append to end of the array
                #  - "-1" set value to last item of the array
                for i in range(len(obj)):
                    key = f"{parent_name}#{sep}"
                    ref_key = f"{parent_name}{i}{sep}"
                    # <-- last item.
                    out_ref_keys[ref_key] = f"{parent_name}-1{sep}"
                    # print("2.3", [key, ref_key])
                    flatten(obj[i], key, ref_key)
        else:
            key = name[:-sep_len]
            if ref_name not in out_keys:
                out_keys[ref_name] = 1
            # print("3.0", [key, ref_name])
            # add to result:
            out.append(KV_Pair(key, obj))

    if data:
        flatten(data)
    # @ret:
    return out



from pprint import pprint

from recursivenamespace import utils
import recursivenamespace


data = {
    "n": 1,
    "t": (2, 3),
    "a": [4, {"c5": "555", "c6": "666", "c7": [1, 2, 3]}],
    "s": set([6, "7", 8]),
    "hello": "hello",
    "d": {"e-1": 9, "e-2": {"e_2_1": 10, "e_2_2": (11,)}, "e_3": [12, 13]},
    "test": [
        {"hello1": "world"},
        {"hello1": "world"},
        {"hello2": "world"},
        {"hello3": {"a": "hello3", "t": [1, 2, 3]}},
        {"hello3": {"a": "hello3", "t": [4, 5, 6]}},
    ],
    "test2": ["1", "2"],
}


def test_flatten_as_dict():
    out = utils.flatten_as_dict(data, flat_list=True)
    print("utils flatten json:")
    pprint(out)

    assert out["a[].1.c7[].2"] == 3


def test_flatten_as_list_0():
    out = utils.flatten_as_list(data)
    print("utils flatten json as list:")
    pprint(out)

    assert out[2][0] == "a"
    assert out[2][1][0] == 4


def test_flatten_as_list_1():
    out = utils.flatten_as_list(
        data, flat_list_type=utils.FlatListType.WITH_INDEX
    )
    print("utils flatten json as list WITH_INDEX:")
    pprint(out)

    assert out[5][0] == "a[].1.c7[].0"
    assert out[5][1] == 1


def test_flatten_as_list_2():
    out = utils.flatten_as_list(
        data, flat_list_type=utils.FlatListType.WITHOUT_INDEX
    )
    print("utils flatten json as list WITHOUT_INDEX:")
    pprint(out)

    assert out[6][0] == "a[].#.c7[].#"
    assert out[6][1] == 2


def test_flatten_as_list_3():
    out = utils.flatten_as_list(
        data, flat_list_type=utils.FlatListType.WITH_SMART_INDEX
    )
    print("utils flatten json as list WITH_SMART_INDEX:")
    # pprint(out)

    assert out[6][0] == "a[].-1.c7[].#"
    assert out[6][1] == 2

    assert out[22][0] == "test[].#.hello3.a"
    assert out[22][1] == "hello3"

    assert out[23][0] == "test[].-1.hello3.t[].#"
    assert out[23][1] == 4


# def main():
#     print("Hello World!")
#     # test_flatten_as_list_3()
#     print(recursivenamespace.__version__)


# if __name__ == "__main__":
#     main()
#     print("..::DONE::..")



#############################
# Fork: https://github.com/HessamLa/recursivenamespace
# %%
import dataclasses
import logging
import re
import sys
from typing import Any, List
from copy import deepcopy
from types import SimpleNamespace
import functools

from . import utils


__all__ = ["recursivenamespace"]


class SetChainKeyError(KeyError):
    def __init__(self, obj, key, sub_key):
        super().__init__(
            f"The object '{key}' typeof({type(obj)}) does not support set[] operator on chain-key '{sub_key}'."
        )


class GetChainKeyError(KeyError):
    def __init__(self, obj, key, sub_key):
        super().__init__(
            f"The object '{key}' typeof({type(obj)}) does not support get[] operator on chain-key '{sub_key}'."
        )


class recursivenamespace(SimpleNamespace):
    __HASH__ = "#"
    __logger = logging.getLogger(__name__)

    def __init__(
        self, data={}, accepted_iter_types=[], use_raw_key=False, **kwargs
    ):
        self.__key_ = ""
        self.__use_raw_key_ = use_raw_key
        self.__supported_types_ = list(
            dict.fromkeys([list, tuple, set] + accepted_iter_types)
        )

        self.__protected_keys_ = ()  # this to add the attr to __dict__.
        self.__protected_keys_ = set(self.__dict__.keys())

        if isinstance(data, dict):
            kwargs.update(data)

        for key, val in kwargs.items():
            key = self.__re(key)
            if isinstance(val, dict):
                val = recursivenamespace(val, accepted_iter_types, use_raw_key)
                val.set_key(key)
            elif isinstance(val, recursivenamespace):
                val.set_key(key)
            else:
                val = self.__process(val)
            # setattr(self, key, val)
            self[key] = val

    def __process(self, val, accepted_iter_types=[], use_raw_key=False):
        if isinstance(val, dict):
            return recursivenamespace(val, accepted_iter_types, use_raw_key)
        elif isinstance(val, str):
            return val
        elif hasattr(val, "__iter__") and type(val) in self.__supported_types_:
            lst = [
                self.__process(v, accepted_iter_types, use_raw_key) for v in val
            ]
            try:
                return type(val)(
                    lst
                )  # the type is assumed to support list-to-type conversion
            except Exception as e:
                print(
                    f"Failed to make iterable object of type {type(val)}",
                    e,
                    out=sys.stderr,
                )
                return val
        else:
            return val

    def __re(self, key):
        return key if self.__use_raw_key_ else re.sub(r"[.\-\s]", "_", key)

    def set_key(self, key):
        self.__key_ = self.__re(key)

    def get_key(self):
        return self.__key_

    def update(self, data):
        try:
            if not isinstance(data, recursivenamespace):
                data = recursivenamespace(
                    data, self.__supported_types_, self.__use_raw_key_
                )
        except:
            raise Exception(f"Failed to update with data of type {type(data)}")
        for key, val in data.items():
            self[key] = val

    def __remove_protected_key(self, key):
        """Use with be-careful!"""
        self.__protected_keys_.remove(key)
        self.__dict__.pop(key)

    def __eq__(self, other):
        if isinstance(other, recursivenamespace):
            return vars(self) == vars(other)
        elif isinstance(other, dict):
            return vars(self) == other
        return False

    def __repr__(self) -> str:
        s = ""
        for k, v in self.items():
            s += f"{k}={v}, "
        if len(s) > 0:
            s = s[:-2]  # remove the last ','
        s = f"RNS({s})"
        return s

    def __str__(self) -> str:
        return self.__repr__()

    def __len__(self):
        return len(self.__dict__) - len(self.__protected_keys_)

    def __delattr__(self, key):
        key = self.__re(key)
        if key not in self.__protected_keys_:
            # delattr(self, key)
            del self.__dict__[key]

    def __setitem__(self, key: str, value: Any):
        key = self.__re(key)
        if key in self.__protected_keys_:
            raise KeyError(f"The key '{key}' is protected.")
        setattr(self, key, value)

    def __getitem__(self, key):
        key = self.__re(key)
        if key in self.__protected_keys_:
            raise KeyError(f"The key '{key}' is protected.")
        return getattr(self, key)

    def __delitem__(self, key):
        key = self.__re(key)
        delattr(self, key)

    def __contains__(self, key):
        key = self.__re(key)
        return key in self.__dict__

    def __copy__(self):
        cls = self.__class__
        result = cls.__new__(cls)
        result.__dict__.update(self.__dict__)
        return result

    def __deepcopy__(self, memo):
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            setattr(result, k, deepcopy(v, memo))
        return result

    def copy(self):
        return self.__copy__()

    def deepcopy(self):
        return self.__deepcopy__()

    def pop(self, key, default=None):
        key = self.__re(key)
        if key in self.__protected_keys_:
            raise KeyError(f"The key '{key}' is protected.")
        if key in self.__dict__:
            val = self.__dict__[key]
            del self.__dict__[key]
            return val
        else:
            return default

    def items(self):
        return [
            (k, v)
            for k, v in self.__dict__.items()
            if k not in self.__protected_keys_
        ]

    def keys(self):
        return [
            k for k in self.__dict__.keys() if k not in self.__protected_keys_
        ]

    def values(self):
        return [
            v
            for k, v in self.__dict__.items()
            if k not in self.__protected_keys_
        ]

    def __iter__(self):
        if sys._getframe(1).f_code.co_name == "dict":
            return self.to_dict()
        return iter(self.keys())

    def to_dict(self, flatten_sep: str = False):
        """Convert the recursivenamespace object to a dictionary.
        If flatten_sep is not False, then the keys are flattened using the separator.
        """
        pairs = []
        for k, v in self.items():
            if isinstance(v, recursivenamespace):
                pairs.append((k, v.to_dict()))
            elif isinstance(v, dict):
                pairs.append((k, v))
            elif hasattr(v, "__iter__") and type(v) in self.__supported_types_:
                pairs.append((k, self.__iter_to_dict(v)))
            else:
                pairs.append((k, v))
        d = dict(pairs)
        if flatten_sep:
            d = dict(utils.flatten_as_dict(d, sep=flatten_sep))
        return d

    def __iter_to_dict(self, iterable=None):
        elements = []
        for val in iterable:
            if isinstance(val, recursivenamespace):
                elements.append(val.to_dict())
            elif isinstance(val, dict):
                elements.append(val)
            elif (
                hasattr(val, "__iter__")
                and type(val) in self.__supported_types_
            ):
                elements.append(self.__iter_to_dict(val))
            else:
                elements.append(val)
        return type(iterable)(elements)

    def __chain_set_array(self, key, subs: List[str], value: any):
        # if the `key` not existed, then create it ??
        if not hasattr(self, key):
            self[key] = []
        target = self[key]
        subs_len = len(subs)
        # validate:
        if not isinstance(target, list):
            raise KeyError(
                f"Invalid array key '{key}'. It is required a list, but got {type(target)}"
            )
        if subs_len == 0:
            raise KeyError(
                f"Invalid array key '{key}'. Required the 'index' as well, e.g.: key[].#"
            )
        # get the `index`:
        index = None if subs[0] == self.__HASH__ else int(subs[0])
        # remove the `index` from sub-key:
        subs = subs[1:]
        subs_len -= 1

        if index is None:  # if APPEND the value ??
            if subs_len == 0:
                # 1) if append the value to the target array
                target.append(value)
            else:
                # 2) if chain-append the value to the target array,
                # then create a "new-item" and set the value for sub-key:
                new_item = recursivenamespace(
                    None, self.__supported_types_, self.__use_raw_key_
                )
                sub_key = utils.join_key(subs)
                new_item.val_set(sub_key, value)
                target.append(new_item)
        else:  # if SET the value ??
            if subs_len == 0:
                # 1) if set the value to the target array at "index"
                target[index] = value
            else:
                # 2) if chain-set the value to the target array at "index",
                # then get the value at "index" and set the value for sub-key:
                target = target[index]
                sub_key = utils.join_key(subs)
                if isinstance(target, recursivenamespace):
                    target.val_set(sub_key, value)
                else:
                    raise SetChainKeyError(target, f"{key}[{index}]", sub_key)

    def __chain_set_value(self, key, subs: List[str], value: any):
        # if the `key` not existed, then create it ??
        if not hasattr(self, key):
            self[key] = recursivenamespace(
                None, self.__supported_types_, self.__use_raw_key_
            )
        target = self[key]
        sub_key = utils.join_key(subs)
        if isinstance(target, recursivenamespace):
            target.val_set(sub_key, value)
        else:
            raise SetChainKeyError(target, key, sub_key)

    def val_set(self, key: str, value: Any):
        """Set the value by key.
        Supported "chain-key", e.g.:
            - a.b.c <- set value to the item "c"
            - a.b.c[].<index-i> <- set value to the item at "index-i" of the array "c", same as: c[i] = value
            - a.b.c[].# <- append value to end of the array "c", same as: c.append(value)
            - a.b.c[].<index-i>.x[].<index-j> <- set value to the item at "index-j" of the array "x", same as: c[i].x[j] = value
            - a.b.c[].<index-i>.x[].# <- append value to end of the array "x", same as: c[i].x.append(value)
            - a.b.c[].#.x[].# <- append value to end of the new-array "x", same as:\n
                - new_item = RNS(dict(x=[])
                - new_item.x.append(value)
                - c.append(new_item)
            - ...

        Args:
            key (str): The key to set
            value (Any): The value to set

        Raises:
            KeyError: when try to set a protected value.
            SetChainKeyError: Only support chain-key on RNS type, else raise the error.
        """
        # raw_key = key
        key, *subs = utils.split_key(key)
        key = utils.unescape_key(key)
        subs_len = len(subs)
        is_array = key[-2:] == utils.KEY_ARRAY

        # if not chain-key/array SET ??
        if subs_len == 0 and not is_array:
            self[key] = value
            return

        # SET to an array
        if is_array:
            self.__chain_set_array(key[:-2], subs, value)
        else:  # SET the value
            self.__chain_set_value(key, subs, value)

    def __chain_get_array(self, key, subs: List[str]):
        target = self[key]
        subs_len = len(subs)
        # validate:
        if not isinstance(target, list):
            raise KeyError(
                f"Invalid array key '{key}'. It is required a list, but got {type(target)}"
            )
        if subs_len == 0:
            raise KeyError(
                f"Invalid array key '{key}'. Required the 'index' as well, e.g.: key[].#"
            )
        # get the `index`:
        index = -1 if subs[0] == self.__HASH__ else int(subs[0])
        # remove the `index` from sub-key:
        subs = subs[1:]
        subs_len -= 1

        # if GET the value by `index`
        if subs_len == 0:
            return target[index]

        # @else: GET value of sub-key
        target = target[index]
        sub_key = utils.join_key(subs)
        if isinstance(target, recursivenamespace):
            return target.val_get(sub_key)
        elif subs_len == 1:
            return getattr(target, sub_key)
        else:
            raise GetChainKeyError(target, key, sub_key)

    def __chain_get_value(self, key, subs: List[str]):
        target = self[key]
        sub_key = utils.join_key(subs)
        if isinstance(target, recursivenamespace):
            return target.val_get(sub_key)
        elif len(subs) == 1:
            return getattr(target, sub_key)
        else:
            raise GetChainKeyError(target, key, sub_key)

    def val_get(self, key: str):
        """Get the value by key.
        Supported "chain-key", e.g.:
            - a.b.c <- get the item "c"
            - a.b.c[].<index-i> <- get the item at "index-i" of the array "c"
            - a.b.c[].# <- get the last item of the array "c" (same as: -1)
            - a.b.c[].<index-i>.x[].<index-j> <- get the item at "index-j" of the array "x"
            - ...

        Args:
            key (str): The key to get

        Raises:
            KeyError: when try to get a protected value/or `key` is not existed.
            GetChainKeyError: Only support chain-key on RNS type, else raise the error.

        Returns:
            any: The value if the `key` is existed.
        """
        # raw_key = key
        key, *subs = utils.split_key(key)
        key = utils.unescape_key(key)
        subs_len = len(subs)
        is_array = key[-2:] == utils.KEY_ARRAY

        # if not chain-key/array GET ??
        if subs_len == 0 and not is_array:
            return self[key]

        # GET from an array
        if is_array:
            return self.__chain_get_array(key[:-2], subs)
        # @else: GET the value
        return self.__chain_get_value(key, subs)

    def get_or_else(self, key: str, or_else=None, show_log=False):
        """Get the value by key.
        Supported "chain-key", e.g.: a.b.c

        Args:
            key (str): The key to get

        Returns:
            any: The value if the `key` is existed, else return `None`.
        """
        try:
            return self.val_get(key)
        except:
            # skip the error.
            if show_log:
                self.__logger.warning(f"KeyNotFound - {key}", exc_info=1)
            return or_else

    def as_schema(self, schema_cls, /, **kwargs):
        if not dataclasses.is_dataclass(schema_cls):
            raise TypeError(f"The 'schema_cls' must be a DataClass type.")
        # @else:
        fields = dataclasses.fields(schema_cls)
        for field in fields:
            name = field.name
            kwargs[name] = self[name]
        return schema_cls(**kwargs)


# %%
def rns(
    accepted_iter_types=[],
    use_raw_key=False,
    use_chain_key=False,
    props="props",
):
    """Create RNS object"""

    def fn_wrapper(func):
        @functools.wraps(func)
        def create_rns(*args, **kwargs):
            # Do something before:
            ret_val = func(*args, **kwargs)

            # Prepare data:
            # create from kv_pair ??
            if (
                use_chain_key
                and isinstance(ret_val, list)
                and (len(ret_val) == 0 or isinstance(ret_val[0], utils.KV_Pair))
            ):
                data = ret_val
            elif isinstance(ret_val, dict):
                data = ret_val
            elif dataclasses.is_dataclass(ret_val):
                data = dataclasses.asdict(ret_val)
            else:
                data = {f"{props}": ret_val}

            # Do something after:
            if use_chain_key:
                ret = recursivenamespace(None, accepted_iter_types, use_raw_key)
                items = data.items() if isinstance(data, dict) else data
                for key, value in items:
                    ret.val_set(key, value)
                return ret
            else:
                return recursivenamespace(
                    data, accepted_iter_types, use_raw_key
                )

        return create_rns

    # @ret:
    return fn_wrapper



import unittest
from recursivenamespace import recursivenamespace as rn


class TestRecursiveNamespace(unittest.TestCase):
    def test0(self):
        data = {"k1": "value1", "k-2": {"k3": "value3", "k4": [100, "value4"]}}
        rn1 = rn(data)
        rn2 = rn(k1="value1", k_2=rn({"k3": "value3", "k4": [100, "value4"]}))
        self.assertEqual(rn1, rn2)
        print(rn1)

    def test_basic_functionality(self):
        data = {"k1": "v1", "k-2": {"k3": "v3", "k4": [100, "v4"]}}
        r = rn(data)
        self.assertEqual(r.k1, "v1")
        self.assertEqual(r.k_2, r["k-2"])
        self.assertEqual(r.k_2, r["k_2"])
        self.assertIsInstance(r.k_2, rn)
        self.assertEqual(r.k_2.k3, "v3")

    def test_equal(self):
        # Test basic functionality of recursivenamespace
        print(
            "rname1 = recursivenamespace(a=1, b=2,c=recursivenamespace(d=3, e=4))"
        )
        rname1 = rn(a=1, b=2, c=rn(d=3, e="404"))
        print(
            "rname2 = recursivenamespace({'a':1, 'b':2, 'c':{'d':3, 'e':'404'}})"
        )
        rname2 = rn({"a": 1, "b": 2, "c": {"d": 3, "e": "404"}})
        print(rname1)
        print(rname2)
        assert rname1 == rname2

    def test_to_dict(self):
        # Test flattening of nested namespaces
        data = {"key1": {"key2": {"key3": "value3", "k4": "v4"}}}
        rn0 = rn(data)
        d = rn0.to_dict()  # Assuming you have a flatten method
        print(d)
        self.assertEqual(d, data)

    def test_to_dict_flatten(self):
        # Test flattening of nested namespaces
        data = {"key1": {"key2": {"key3": "value3", "k4": ["v4", "v5"]}}}
        rn0 = rn(data)
        d = rn0.to_dict(flatten_sep="_")
        self.assertEqual(
            d, {"key1_key2_key3": "value3", "key1_key2_k4": ["v4", "v5"]}
        )

    # Add more tests as necessary for your class functionalities


# if __name__ == "__main__":
#     unittest.main()



from .main import recursivenamespace
from .main import recursivenamespace as RecursiveNamespace
from .main import recursivenamespace as RNS
from . import main as rns

try:
    from recursivenamespace._version_pdm import (  # pyright: ignore [reportMissingImports]
        __version__,
    )
except ImportError:
    from recursivenamespace._version import get_versions

    __version__ = get_versions()["version"]
    del get_versions

__all__ = ["recursivenamespace", "RecursiveNamespace", "RNS", "rns"]



from pathlib import Path
import pickle
from pprint import pprint
import dataclasses

from recursivenamespace import rns
from recursivenamespace import RecursiveNamespace
from recursivenamespace import utils


PJ_DIR = Path(__file__).parents[1]


data = {
    "n": 1,
    "t": (2, 3),
    "a": [4, [1, 2], {"c5": "555", "c6": "666", "c7": [1, 2, 3]}],
    "s": set([6, "7", 8]),
    "hello": "hello",
    "d": {"e-1": 9, "e-2": {"e_2_1": 10, "e_2_2": (11,)}, "e_3": [12, 13]},
    "v1": "Hello",
    "v2": "World",
}
rns_1 = RecursiveNamespace(data, use_raw_key=False)
rns_2 = RecursiveNamespace(data, use_raw_key=True)


@rns.rns()
@dataclasses.dataclass
class FreeItem:
    v1: str
    v2: int


@dataclasses.dataclass
class SchemaItem:
    v1: str
    v2: int


@rns.rns(use_chain_key=True)
def create_by_chain_key():
    return {"x.y.z": [1, 2, 3], "s": "hello world"}


def test_convert():
    d = dict(rns_1)
    print("dict(rns_1)")
    print(d)
    print("rns_1")
    pprint(rns_1)
    print("rns_2")
    pprint(rns_2)
    # assert:
    assert rns_1.n == d["n"]
    assert rns_1.d.e_1 == d["d"]["e_1"]
    assert rns_2.d["e-1"] == data["d"]["e-1"]


def test_pickle():
    pklpath = PJ_DIR / "out" / "test.pkl"
    if not pklpath.parent.exists():
        pklpath.parent.mkdir()

    # pickel test
    with open(pklpath, "wb") as f:
        pickle.dump(rns_2, f)
    with open(pklpath, "rb") as f:
        rname = pickle.load(f)

    print()
    print("Result after unpickling")
    print(rname)

    assert rns_2 == rname


def test_decorator():
    a = FreeItem("Hello", 1)
    pprint(a)
    assert a.v1 == "Hello"
    assert isinstance(a, RecursiveNamespace)


def test_schema():
    a: SchemaItem = rns_1.as_schema(SchemaItem)
    pprint(a)
    assert a.v1 == "Hello"
    assert a.v2 == "World"
    assert isinstance(a, SchemaItem)


def test_flattern():
    print("RNS object:")
    print(rns_2)
    # 1:
    print("Flatten - RNS")
    pprint(rns_2.to_dict(flatten_sep="."))
    # 2:
    print("Flatten - Json")
    pprint(dict(utils.flatten_as_dict(rns_2.to_dict())))


def test_get_set():
    rns_2["x.y"] = "single-key"
    rns_2.val_set("x.y", "chain-key")
    rns_2.val_set("a[].0", -4)
    rns_2.val_set("a[].1", 100)
    rns_2.val_set("a[].-1.c7[].-1", 10)
    rns_2.val_set("a[].-1.c7[].#", 100)
    rns_2.val_set("a[].#.c7[].#", 100)
    rns_2.val_set("a[].-1.c8", dict(x=123, y="456"))
    rns_2.val_set("a[].-1.c9.c10", dict(x=123, y="456"))
    rns_2.val_set("a[].-1.c9.c11", [dict(x=123, y="456")])
    rns_2.val_set("a[].-1.c9.c11[].#", "a string")

    # debug:
    print(f"===>>>>>> Test case: {test_get_set.__name__}")
    pprint(rns_2)

    assert rns_2["x.y"] == "single-key"
    assert rns_2.val_get(r"x\\.y") == "single-key"
    assert rns_2.val_get("x.y") == "chain-key"
    assert rns_2.val_get("a[].-2.c7[].-2") == 10
    assert rns_2.val_get("a[].-2.c7[].-1") == 100
    assert rns_2.val_get("a[].-1.c7[].-1") == 100
    # assert rns_2.get_or_none("a[].#.c7[].#") == 100


# def main():
#     # print(create_by_chain_key())
#     test_get_set()
#     print("Hello World!")


# if __name__ == "__main__":
#     main()
#     print("..::DONE::..")
